@@ -28,7 +28,9 @@
 #include <syslog.h>
 #include <stdarg.h>
 #include <signal.h>
+#ifdef __linux__
 #include <sys/prctl.h>
+#endif
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/resource.h>
@@ -52,8 +54,6 @@
 #endif
 
 /* Module defines */
-#define LINE_LENGTH 1024
-
 #define LIMITS_DEF_USER     0 /* limit was set by a user entry */
 #define LIMITS_DEF_GROUP    1 /* limit was set by a group entry */
 #define LIMITS_DEF_ALLGROUP 2 /* limit was set by a group entry */
@@ -97,7 +97,7 @@ struct pam_limit_s {
     struct user_limits_struct limits[RLIM_NLIMITS];
     const char *conf_file;
     int utmp_after_pam_call;
-    char login_group[LINE_LENGTH];
+    char *login_group;
 };
 
 #define LIMIT_LOGIN RLIM_NLIMITS+1
@@ -293,6 +293,7 @@ check_logins (pam_handle_t *pamh, const char *name, int limit, int ctrl,
                 continue;
 	    }
 	    if ((pl->login_limit_def == LIMITS_DEF_ALLGROUP)
+		&& pl->login_group != NULL
 		&& !pam_modutil_user_in_group_nam_nam(pamh, user, pl->login_group)) {
                 continue;
 	    }
@@ -321,6 +322,7 @@ check_logins (pam_handle_t *pamh, const char *name, int limit, int ctrl,
     return 0;
 }
 
+#ifdef __linux__
 static const char *lnames[RLIM_NLIMITS] = {
         [RLIMIT_CPU] = "Max cpu time",
         [RLIMIT_FSIZE] = "Max file size",
@@ -392,10 +394,11 @@ static rlim_t str2rlim_t(char *value) {
 
 static void parse_kernel_limits(pam_handle_t *pamh, struct pam_limit_s *pl, int ctrl)
 {
-    int i, maxlen = 0;
+    int i;
     FILE *limitsfile;
     const char *proclimits = "/proc/1/limits";
-    char line[256];
+    char *line = NULL;
+    size_t maxlen = 0, n = 0;
     char *hard, *soft, *name;
 
     if (!(limitsfile = fopen(proclimits, "r"))) {
@@ -403,8 +406,8 @@ static void parse_kernel_limits(pam_handle_t *pamh, struct pam_limit_s *pl, int
         return;
     }
 
-    while (fgets(line, 256, limitsfile)) {
-        int pos = strlen(line);
+    while (getline(&line, &n, limitsfile) != -1) {
+        size_t pos = strlen(line);
         if (pos < 2) continue;
 
         /* drop trailing newline */
@@ -448,8 +451,10 @@ static void parse_kernel_limits(pam_handle_t *pamh, struct pam_limit_s *pl, int
         pl->limits[i].src_soft = LIMITS_DEF_KERNEL;
         pl->limits[i].src_hard = LIMITS_DEF_KERNEL;
     }
+    free(line);
     fclose(limitsfile);
 }
+#endif
 
 static int init_limits(pam_handle_t *pamh, struct pam_limit_s *pl, int ctrl)
 {
@@ -492,6 +497,7 @@ static int init_limits(pam_handle_t *pamh, struct pam_limit_s *pl, int ctrl)
       retval = !PAM_SUCCESS;
     pl->login_limit = -2;
     pl->login_limit_def = LIMITS_DEF_NONE;
+    pl->login_group = NULL;
 
     return retval;
 }
@@ -504,14 +510,16 @@ static int init_limits(pam_handle_t *pamh, struct pam_limit_s *pl, int ctrl)
 static int
 value_from_file(const char *pathname, rlim_t *valuep)
 {
-    char buf[128];
     FILE *fp;
     int retval;
 
     retval = 0;
 
     if ((fp = fopen(pathname, "r")) != NULL) {
-	if (fgets(buf, sizeof(buf), fp) != NULL) {
+	char *buf = NULL;
+	size_t n = 0;
+
+	if (getline(&buf, &n, fp) != -1) {
 	    char *endptr;
 	    unsigned long long value;
 
@@ -525,6 +533,7 @@ value_from_file(const char *pathname, rlim_t *valuep)
 	    }
 	}
 
+	free(buf);
 	fclose(fp);
     }
 
@@ -813,12 +822,46 @@ parse_uid_range(pam_handle_t *pamh, const char *domain,
     return rv;
 }
 
+static int
+set_if_null(char **dest, char *def)
+{
+    if (*dest == NULL) {
+	*dest = def;
+	return 0;
+    }
+    return 1;
+}
+
+static int
+split(char *line, char **domain, char **ltype, char **item, char **value)
+{
+    char *blank, *saveptr;
+    int count;
+
+    blank = line + strlen(line);
+    saveptr = NULL;
+
+    *domain = strtok_r(line, " \t", &saveptr);
+    *ltype = strtok_r(NULL, " \t", &saveptr);
+    *item = strtok_r(NULL, " \t", &saveptr);
+    *value = strtok_r(NULL, "", &saveptr);
+
+    count = 0;
+    count += set_if_null(domain, blank);
+    count += set_if_null(ltype, blank);
+    count += set_if_null(item, blank);
+    count += set_if_null(value, blank);
+
+    return count;
+}
+
 static int
 parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
 		  int ctrl, struct pam_limit_s *pl, const int conf_file_set_by_user)
 {
     FILE *fil;
-    char buf[LINE_LENGTH];
+    char *buf = NULL;
+    size_t n = 0;
 
     /* check for the conf_file */
     if (ctrl & PAM_DEBUG_ARG)
@@ -835,20 +878,16 @@ parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
     }
 
     /* start the show */
-    while (fgets(buf, LINE_LENGTH, fil) != NULL) {
-        char domain[LINE_LENGTH];
-        char ltype[LINE_LENGTH];
-        char item[LINE_LENGTH];
-        char value[LINE_LENGTH];
+    while (getline(&buf, &n, fil) != -1) {
+        char *domain, *ltype, *item, *value, *tptr, *line;
         int i;
         int rngtype;
         size_t j;
-        char *tptr,*line;
         uid_t min_uid = (uid_t)-1, max_uid = (uid_t)-1;
 
         line = buf;
         /* skip the leading white space */
-        while (*line && isspace(*line))
+        while (*line && isspace((unsigned char)*line))
             line++;
 
         /* Rip off the comments */
@@ -863,14 +902,12 @@ parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
         if (!strlen(line))
             continue;
 
-	domain[0] = ltype[0] = item[0] = value[0] = '\0';
-
-	i = sscanf(line,"%s%s%s%s", domain, ltype, item, value);
+	i = split(line, &domain, &ltype, &item, &value);
 	D(("scanned line[%d]: domain[%s], ltype[%s], item[%s], value[%s]",
 	   i, domain, ltype, item, value));
 
         for(j=0; j < strlen(ltype); j++)
-            ltype[j]=tolower(ltype[j]);
+            ltype[j]=tolower((unsigned char)ltype[j]);
 
 	if ((rngtype=parse_uid_range(pamh, domain, &min_uid, &max_uid)) < 0) {
 	    pam_syslog(pamh, LOG_WARNING, "invalid uid range '%s' - skipped", domain);
@@ -879,11 +916,11 @@ parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
 
         if (i == 4) { /* a complete line */
 	    for(j=0; j < strlen(item); j++)
-		item[j]=tolower(item[j]);
+		item[j]=tolower((unsigned char)item[j]);
 	    for(j=0; j < strlen(value); j++)
-		value[j]=tolower(value[j]);
+		value[j]=tolower((unsigned char)value[j]);
 
-            if (strcmp(uname, domain) == 0) /* this user have a limit */
+            if (strcmp(uname, domain) == 0) /* this user has a limit */
                 process_limit(pamh, LIMITS_DEF_USER, ltype, item, value, ctrl, pl);
             else if (domain[0]=='@') {
 		if (ctrl & PAM_DEBUG_ARG) {
@@ -923,7 +960,8 @@ parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
 			    process_limit(pamh, LIMITS_DEF_ALL, ltype, item, value, ctrl,
 					  pl);
 			else if (pam_modutil_user_in_group_nam_nam(pamh, uname, domain+1)) {
-			    strcpy(pl->login_group, domain+1);
+			    free(pl->login_group);
+			    pl->login_group = strdup(domain+1);
 			    process_limit(pamh, LIMITS_DEF_ALLGROUP, ltype, item, value, ctrl,
 					  pl);
 			}
@@ -932,8 +970,8 @@ parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
 			if (pam_modutil_user_in_group_nam_gid(pamh, uname, (gid_t)max_uid)) {
 			    struct group *grp;
 			    grp = pam_modutil_getgrgid(pamh, (gid_t)max_uid);
-			    strncpy(pl->login_group, grp->gr_name, sizeof(pl->login_group));
-			    pl->login_group[sizeof(pl->login_group)-1] = '\0';
+			    free(pl->login_group);
+			    pl->login_group = strdup(grp->gr_name);
 			    process_limit(pamh, LIMITS_DEF_ALLGROUP, ltype, item, value, ctrl,
 					  pl);
 			}
@@ -1011,12 +1049,14 @@ parse_config_file(pam_handle_t *pamh, const char *uname, uid_t uid, gid_t gid,
 		    pam_syslog(pamh, LOG_DEBUG, "no limits for '%s'", uname);
 		}
 	    }
+	    free(buf);
 	    fclose(fil);
 	    return PAM_IGNORE;
         } else {
             pam_syslog(pamh, LOG_WARNING, "invalid line '%s' - skipped", line);
 	}
     }
+    free(buf);
     fclose(fil);
     return PAM_SUCCESS;
 }
@@ -1078,16 +1118,20 @@ static int setup_limits(pam_handle_t *pamh,
     }
 
     if (pl->nonewprivs) {
+#ifdef __linux__
 	if (prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0) < 0) {
 	    pam_syslog(pamh, LOG_ERR, "Could not set prctl(PR_SET_NO_NEW_PRIVS): %m");
 	    retval |= LIMIT_ERR;
 	}
+#else
+	pam_syslog(pamh, LOG_INFO, "Setting 'nonewprivs' not supported on this OS");
+#endif
     }
 
     return retval;
 }
 
-/* --- evaluting all files in VENDORDIR/security/limits.d and /etc/security/limits.d --- */
+/* --- evaluating all files in VENDORDIR/security/limits.d and /etc/security/limits.d --- */
 static const char *
 base_name(const char *path)
 {
@@ -1232,6 +1276,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 			       ctrl, pl, conf_file_set_by_user);
     if (retval == PAM_IGNORE) {
 	D(("the configuration file ('%s') has an applicable '<domain> -' entry", pl->conf_file));
+	free(pl->login_group);
 	return PAM_SUCCESS;
     }
     if (retval != PAM_SUCCESS || conf_file_set_by_user)
@@ -1254,6 +1299,7 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 
     if (retval == PAM_IGNORE) {
         D(("the configuration file ('%s') has an applicable '<domain> -' entry", pl->conf_file));
+        free(pl->login_group);
         return PAM_SUCCESS;
     }
 
@@ -1261,10 +1307,12 @@ out:
     if (retval != PAM_SUCCESS)
     {
 	pam_syslog(pamh, LOG_ERR, "error parsing the configuration file: '%s' ", pl->conf_file);
+	free(pl->login_group);
 	return retval;
     }
 
     retval = setup_limits(pamh, pwd->pw_name, pwd->pw_uid, ctrl, pl);
+    free(pl->login_group);
     if (retval & LOGIN_ERR)
 	pam_error(pamh, _("There were too many logins for '%s'."),
 		  pwd->pw_name);
