@@ -52,6 +52,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
+#include <signal.h>
 
 #include <security/pam_modules.h>
 #include <security/_pam_macros.h>
@@ -72,11 +73,6 @@
 #define XAUTHDEF ".Xauthority"
 #define XAUTHTMP ".xauthXXXXXX"
 
-/* Hurd compatibility */
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
 /* Possible paths to xauth executable */
 static const char * const xauthpaths[] = {
 #ifdef PAM_PATH_XAUTH
@@ -91,14 +87,14 @@ static const char * const xauthpaths[] = {
  * given input on stdin, and storing any output it generates. */
 static int
 run_coprocess(pam_handle_t *pamh, const char *input, char **output,
-	      uid_t uid, gid_t gid, const char *command, ...)
+	      uid_t uid, gid_t gid, const char *command, const char *argv[])
 {
 	int ipipe[2], opipe[2], i;
 	char buf[LINE_MAX];
 	pid_t child;
 	char *buffer = NULL;
 	size_t buffer_size = 0;
-	va_list ap;
+	struct sigaction newsa, oldsa;
 
 	*output = NULL;
 
@@ -114,6 +110,17 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 		return -1;
 	}
 
+	memset(&newsa, '\0', sizeof(newsa));
+	newsa.sa_handler = SIG_DFL;
+	if (sigaction(SIGCHLD, &newsa, &oldsa) == -1) {
+		pam_syslog(pamh, LOG_ERR, "failed to reset SIGCHLD handler: %m");
+		close(ipipe[0]);
+		close(ipipe[1]);
+		close(opipe[0]);
+		close(opipe[1]);
+		return -1;
+	}
+
 	/* Fork off a child. */
 	child = fork();
 	if (child == -1) {
@@ -126,9 +133,6 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 	}
 
 	if (child == 0) {
-		/* We're the child. */
-		size_t j;
-		const char *args[10];
 		/* Drop privileges. */
 		if (setgid(gid) == -1)
 		  {
@@ -168,20 +172,9 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 						    PAM_MODUTIL_NULL_FD) < 0) {
 		    _exit(1);
 		}
-		/* Initialize the argument list. */
-		memset(args, 0, sizeof(args));
-		/* Convert the varargs list into a regular array of strings. */
-		va_start(ap, command);
-		args[0] = command;
-		for (j = 1; j < PAM_ARRAY_SIZE(args) - 1; j++) {
-			args[j] = va_arg(ap, const char*);
-			if (args[j] == NULL) {
-				break;
-			}
-		}
 		/* Run the command. */
 		DIAG_PUSH_IGNORE_CAST_QUAL;
-		execv(command, (char *const *) args);
+		execv(command, (char *const *) argv);
 		DIAG_POP_IGNORE_CAST_QUAL;
 		/* Never reached. */
 		_exit(1);
@@ -204,11 +197,10 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 		tmp = realloc(buffer, buffer_size + i + 1);
 		if (tmp == NULL) {
 			/* Uh-oh, bail. */
-			if (buffer != NULL) {
-				free(buffer);
-			}
+			free(buffer);
 			close(opipe[0]);
 			waitpid(child, NULL, 0);
+			sigaction(SIGCHLD, &oldsa, NULL);   /* restore old signal handler */
 			return -1;
 		}
 		/* Save the new buffer location, copy the newly-read data into
@@ -225,6 +217,7 @@ run_coprocess(pam_handle_t *pamh, const char *input, char **output,
 	close(opipe[0]);
 	*output = buffer;
 	waitpid(child, NULL, 0);
+	sigaction(SIGCHLD, &oldsa, NULL);   /* restore old signal handler */
 	return 0;
 }
 
@@ -242,7 +235,7 @@ check_acl(pam_handle_t *pamh,
 	  const char *sense, const char *this_user, const char *other_user,
 	  int noent_code, int debug)
 {
-	char path[PATH_MAX];
+	char *path = NULL;
 	struct passwd *pwd;
 	FILE *fp = NULL;
 	int i, fd = -1, save_errno;
@@ -258,14 +251,17 @@ check_acl(pam_handle_t *pamh,
 		return PAM_SESSION_ERR;
 	}
 	/* Figure out what that file is really named. */
-	i = snprintf(path, sizeof(path), "%s/.xauth/%s", pwd->pw_dir, sense);
-	if ((i >= (int)sizeof(path)) || (i < 0)) {
+	i = asprintf(&path, "%s/.xauth/%s", pwd->pw_dir, sense);
+	if (i < 0) {
 		pam_syslog(pamh, LOG_ERR,
-			   "name of user's home directory is too long");
+			   "cannot allocate path buffer for ~/.xauth/%s",
+			   sense);
 		return PAM_SESSION_ERR;
 	}
-	if (pam_modutil_drop_priv(pamh, &privs, pwd))
+	if (pam_modutil_drop_priv(pamh, &privs, pwd)) {
+		free(path);
 		return PAM_SESSION_ERR;
+	}
 	if (!stat(path, &st)) {
 		if (!S_ISREG(st.st_mode))
 			errno = EINVAL;
@@ -276,6 +272,7 @@ check_acl(pam_handle_t *pamh,
 	if (pam_modutil_regain_priv(pamh, &privs)) {
 		if (fd >= 0)
 			close(fd);
+		free(path);
 		return PAM_SESSION_ERR;
 	}
 	if (fd >= 0) {
@@ -291,24 +288,20 @@ check_acl(pam_handle_t *pamh,
 		}
 	}
 	if (fp) {
-		char buf[LINE_MAX], *tmp;
+		char *buf = NULL;
+		size_t n = 0;
 		/* Scan the file for a list of specs of users to "trust". */
-		while (fgets(buf, sizeof(buf), fp) != NULL) {
-			tmp = memchr(buf, '\r', sizeof(buf));
-			if (tmp != NULL) {
-				*tmp = '\0';
-			}
-			tmp = memchr(buf, '\n', sizeof(buf));
-			if (tmp != NULL) {
-				*tmp = '\0';
-			}
+		while (getline(&buf, &n, fp) != -1) {
+			buf[strcspn(buf, "\r\n")] = '\0';
 			if (fnmatch(buf, other_user, 0) == 0) {
 				if (debug) {
 					pam_syslog(pamh, LOG_DEBUG,
 						   "%s %s allowed by %s",
 						   other_user, sense, path);
 				}
+				free(buf);
 				fclose(fp);
+				free(path);
 				return PAM_SUCCESS;
 			}
 		}
@@ -317,7 +310,9 @@ check_acl(pam_handle_t *pamh,
 			pam_syslog(pamh, LOG_DEBUG, "%s not listed in %s",
 				   other_user, path);
 		}
+		free(buf);
 		fclose(fp);
+		free(path);
 		return PAM_PERM_DENIED;
 	} else {
 		/* Default to okay if the file doesn't exist. */
@@ -337,12 +332,19 @@ check_acl(pam_handle_t *pamh,
 						   path);
 				}
 			}
+			free(path);
 			return noent_code;
+		case ENAMETOOLONG:
+			pam_syslog(pamh, LOG_ERR,
+				   "error opening %s: %m", path);
+			free(path);
+			return PAM_SESSION_ERR;
 		default:
 			if (debug) {
 				pam_syslog(pamh, LOG_DEBUG,
 					   "error opening %s: %m", path);
 			}
+			free(path);
 			return PAM_PERM_DENIED;
 		}
 	}
@@ -502,16 +504,13 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 	/* Figure out where the source user's .Xauthority file is. */
 	if (getenv(XAUTHENV) != NULL) {
 		cookiefile = strdup(getenv(XAUTHENV));
-	} else {
-		cookiefile = malloc(strlen(rpwd->pw_dir) + 1 +
-				    strlen(XAUTHDEF) + 1);
 		if (cookiefile == NULL) {
 			retval = PAM_SESSION_ERR;
 			goto cleanup;
 		}
-		strcpy(cookiefile, rpwd->pw_dir);
-		strcat(cookiefile, "/");
-		strcat(cookiefile, XAUTHDEF);
+	} else if (asprintf(&cookiefile, "%s/%s", rpwd->pw_dir, XAUTHDEF) < 0) {
+		retval = PAM_SESSION_ERR;
+		goto cleanup;
 	}
 	if (debug) {
 		pam_syslog(pamh, LOG_DEBUG, "reading keys from `%s'",
@@ -529,8 +528,9 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 	}
 	if (run_coprocess(pamh, NULL, &cookie,
 			  getuid(), getgid(),
+			  xauth, (const char *[]) {
 			  xauth, "-f", cookiefile, "nlist", display,
-			  NULL) == 0) {
+			  NULL}) == 0) {
 #ifdef WITH_SELINUX
 		char *context_raw = NULL;
 #endif
@@ -540,32 +540,18 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 		if (((cookie == NULL) || (strlen(cookie) == 0)) &&
 		    (pam_str_skip_prefix(display, "localhost:") != NULL ||
 		     pam_str_skip_prefix(display, "localhost/unix:") != NULL)) {
-			char *t, *screen;
-			size_t tlen, slen;
+			char hostname[HOST_NAME_MAX + 1];
 			/* Free the useless cookie string. */
-			if (cookie != NULL) {
-				free(cookie);
-				cookie = NULL;
-			}
-			/* Allocate enough space to hold an adjusted name. */
-			tlen = strlen(display) + LINE_MAX + 1;
-			t = malloc(tlen);
-			if (t != NULL) {
-				memset(t, 0, tlen);
-				if (gethostname(t, tlen - 1) != -1) {
-					/* Append the protocol and then the
-					 * screen number. */
-					if (strlen(t) < tlen - 6) {
-						strcat(t, "/unix:");
-					}
-					screen = strchr(display, ':');
-					if (screen != NULL) {
-						screen++;
-						slen = strlen(screen);
-						if (strlen(t) + slen < tlen) {
-							strcat(t, screen);
-						}
-					}
+			free(cookie);
+			cookie = NULL;
+			if (gethostname(hostname, sizeof(hostname)) != -1) {
+				const char *screen;
+				char *t;
+
+				/* Append protocol and screen number to host. */
+				screen = display + strcspn(display, ":");
+				if (asprintf(&t, "%s/unix%s",
+					     hostname, screen) >= 0) {
 					if (debug) {
 						pam_syslog(pamh, LOG_DEBUG,
 							   "no key for `%s', "
@@ -588,11 +574,11 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 					}
 					run_coprocess(pamh, NULL, &cookie,
 						      getuid(), getgid(),
+						      xauth, (const char *[]) {
 						      xauth, "-f", cookiefile,
-						      "nlist", t, NULL);
+						      "nlist", t, NULL});
+					free(t);
 				}
-				free(t);
-				t = NULL;
 			}
 		}
 
@@ -730,7 +716,8 @@ pam_sm_open_session (pam_handle_t *pamh, int flags UNUSED,
 		}
 		run_coprocess(pamh, cookie, &tmp,
 			      tpwd->pw_uid, tpwd->pw_gid,
-			      xauth, "-f", cookiefile, "nmerge", "-", NULL);
+			      xauth, (const char *[]) {
+			      xauth, "-f", cookiefile, "nmerge", "-", NULL});
 
 		/* We don't need to keep a copy of these around any more. */
 		cookiefile = NULL;
